@@ -1,40 +1,22 @@
 from collections.abc import MutableMapping
 from inspect import isclass
 from itertools import chain
-from typing import Any, Dict, Generic, TypeVar
+from typing import Any, Dict, Generic, TypeVar, Union, Tuple, List, Optional
+from typing import get_type_hints
 
 import dill
 import numpy as np
 from typing_inspect import get_generic_bases, get_origin, get_args
 
 
-class FileBackedAttribute:
-    """Class attribute that is backed by file storage. Usage:
-
-    class MyClass:
-        my_attr = FileBackedAttribute(type)
-
-    This is a meta-object that is decoded at type creation time, and
-    only works for classes derived from FileBacked.
-    """
-
-    def __init__(self, tp=None):
-        self.tp = tp
-
-    def descriptor(self, name):
-        return FileBackedDescriptor(name, self.tp)
-
-
 class FileBackedDescriptor:
     """Descriptor that allows attribute-like access to an object backed by
     file storage.  These objects should not be created manually,
-    instead they are created by FileBackedAttribute when the metaclass
-    requires it.
+    instead they are created the FileBacked metaclass when required.
     """
 
-    def __init__(self, name, tp):
+    def __init__(self, name):
         self.name = name
-        self.tp = tp
 
     def __get__(self, obj, owner=None):
         if obj is None:
@@ -53,11 +35,13 @@ class FileBackedDescriptor:
     def write(self, group, obj, **kwargs):
         """Issue a call to write this object to a HDF5 group."""
         value = self.__get__(obj)
-        write(group, self.name, value, self.tp, **kwargs)
+        tp = obj.__filebacked_annotations__[self.name]
+        write(group, self.name, value, tp, **kwargs)
 
     def read(self, obj, lazy=False):
         """Issue a call to read this object from a HDF5 group."""
-        return _getkey(obj, self.name, self.tp, AttributeError, lazy)
+        tp = obj.__filebacked_annotations__[self.name]
+        return _getkey(obj, self.name, tp, AttributeError, lazy)
 
 
 def _getkey(obj, key, tp, exc, lazy, filekey=None):
@@ -91,6 +75,10 @@ def _getkey(obj, key, tp, exc, lazy, filekey=None):
         obj.__filebacked_data__[key] = value
         return value
 
+    # If the type is of option type, it's missing
+    elif _is_option(tp):
+        return None
+
     raise exc(key)
 
 
@@ -168,6 +156,7 @@ class FileBackedBase:
         self.__filebacked_group__ = group
         self.__filebacked_data__ = {}
         self.__filebacked_deleted__ = set()
+        self.__filebacked_annotations__ = get_type_hints(self.__class__)
 
         if call_pyinit:
             self.__pyinit__()
@@ -239,9 +228,15 @@ class FileBackedMeta(type):
 
     def __new__(cls, name, bases, attrs):
         # Collect all file-backed attributes from this class and convert them to descriptors
-        file_attribs = {attr for attr, obj in attrs.items() if isinstance(obj, FileBackedAttribute)}
-        for attr in file_attribs:
-            attrs[attr] = attrs[attr].descriptor(attr)
+        ignore = attrs.get('__filebacked_ignore__', set())
+        annotations = attrs.get('__annotations__', dict())
+
+        file_attribs = set()
+        for name in annotations:
+            if name in ignore:
+                continue
+            file_attribs.add(name)
+            attrs[name] = FileBackedDescriptor(name)
 
         # Track all file-backed attributes from the entire inheritance diagram
         for base in bases:
@@ -459,9 +454,9 @@ class BuiltinSequenceFilter:
 
     def applicable(self, tp):
         orig = get_origin(tp)
-        if orig == list:
+        if orig == list or orig == List:
             return True
-        if orig != tuple:
+        if orig != tuple and orig != Tuple:
             return False
         args = get_args(tp)
         return len(args) == 2 and args[1] == Ellipsis
@@ -477,28 +472,48 @@ class BuiltinSequenceFilter:
 
     def read(self, group, tp, **kwargs):
         eltype, *_ = get_args(tp)
+        constructor = get_origin(tp)
+        if constructor == Tuple:
+            constructor = tuple
+        elif constructor == List:
+            constructor = list
         if _is_scalar(eltype):
-            return get_origin(tp)(group[:])
-        return get_origin(tp)(read(group[str(i)], eltype, **kwargs) for i in range(len(group)))
+            return constructor(group[:])
+        return constructor(read(group[str(i)], eltype, **kwargs) for i in range(len(group)))
 
 
 class DictFilter:
     """Filter for dictionaries."""
 
     def applicable(self, tp):
-        return get_origin(tp) == dict
+        return get_origin(tp) == dict or get_origin(tp) == Dict
 
     def write(self, group, name, obj, tp, **kwargs):
-        K, V = get_args(tp)
+        K, V = get_args(tp, evaluate=True)
         _write_dict(group.require_group(name), obj, K, V, **kwargs)
 
     def read(self, group, tp, **kwargs):
-        K, V = get_args(tp)
+        K, V = get_args(tp, evaluate=True)
         retval = {}
         _read_dict(group, retval, K, V, **kwargs)
         return retval
 
 
+class OptionFilter:
+    """Filter for Option[T] types."""
+
+    def applicable(self, tp):
+        return _is_option(tp)
+
+    def write(self, group, name, obj, tp, **kwargs):
+        if obj is not None:
+            write(group, name, obj, **kwargs)
+
+    def read(self, group, tp, **kwargs):
+        T, _ = get_args(tp)
+        return read(group, T, **kwargs)
+
+
 class PickleFilter:
     """Filter for I/O of arbitrary Python objects, stored as pickled
     strings.
@@ -520,6 +535,7 @@ _FILTERS = [
     ScalarFilter(),
     NumpyFilter(),
     BuiltinSequenceFilter(),
+    OptionFilter(),
     DictFilter(),
 ]
 
@@ -550,7 +566,7 @@ def write(group, name, obj, tp=None, **kwargs):
         if flt.applicable(tp):
             flt.write(group, name, obj, tp, **kwargs)
             return
-    if allow_pickle:
+    if allow_pickle or tp == object:
         pickle_filter.write(group, name, obj, tp, **kwargs)
         return
     raise TypeError(f"Unable to write type '{tp}'")
@@ -570,7 +586,7 @@ def read(group, tp, **kwargs):
     for flt in _FILTERS:
         if flt.applicable(tp):
             return flt.read(group, tp, **kwargs)
-    if allow_pickle:
+    if allow_pickle or tp == object:
         return pickle_filter.read(group, tp, **kwargs)
     raise TypeError(f"Unable to read type '{tp}'")
 
@@ -654,7 +670,7 @@ def _my_typevars(obj, cls):
         obj = type(obj)
     for base in get_generic_bases(obj):
         if get_origin(base) == cls:
-            return get_args(base)
+            return get_args(base, evaluate=True)
     raise TypeError(f"Couldn't find type variables for {obj} relative to {cls}")
 
 
@@ -666,7 +682,7 @@ def _subclasses(cls, root=False):
         yield cls
     for sub in cls.__subclasses__():
         yield sub
-        yield from subclasses(sub, root=False)
+        yield from _subclasses(sub, root=False)
 
 
 def _distill_attribs(attribs, only, skip):
@@ -684,10 +700,21 @@ def _distill_attribs(attribs, only, skip):
 
 
 _SCALARS = {
-    float, np.float, np.float128, np.float64, np.float32, np.float16,
+    float, np.float, np.float64, np.float32, np.float16,
     int, np.int, np.int64, np.int32, np.int16, np.int8, np.int0,
 }
 
+if hasattr(np, 'float128'):
+    _SCALARS.add(np.float128)
+
 def _is_scalar(tp):
     """Check if a type is considered a scalar."""
     return tp in _SCALARS
+
+def _is_option(tp):
+    """Return true if tp is an option type."""
+    orig = get_origin(tp)
+    if orig != Union:
+        return False
+    args = get_args(tp)
+    return args[1] == type(None)
